from __future__ import annotations

import abc
import collections.abc as tabc
import typing as typ
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from threading import RLock

P = typ.ParamSpec("P")
T = typ.TypeVar("T")
RT = typ.TypeVar("RT")

RootType = typ.NewType("RootType", tabc.Mapping)
"""
:py:class:`~typing.NewType` used to type the configuration root.

Aliases :py:class:`~collections.abc.Mapping` as root has to be a mapping for it to be used and no Tag should mutate it.
"""

Root = RootType | None
"""
:py:data:`~typing.TypeAlias` used by type checking to identify the configuration root if it exists.
"""

Tag = typ.NewType("Tag", str)
"""
:py:class:`~typing.NewType` used to type tag strings. Must begin with ``!``.
"""


class Masked(str):
    """
    Used to keep secrets from printing to screen when running tests.

    - Inherits from :py:class:`str`.
    - Replaces the standard :py:meth:`~object.__repr__` result with the
      constant literal ``'<****>'``.
    - Used by ``!Mask`` tag

    Note:
        Does not alter text or prevent :py:func:`print` from display the string value.
    """

    def __repr__(self) -> str:
        return "'<****>'"


class Placeholder:
    """
    Representation of ``!Placeholder`` tag.

    Holds the ``!Placeholder`` message.
    """

    __slot__ = ("message",)

    def __init__(self, message: str) -> None:
        self.message: typ.Final = message

    def __str__(self) -> str:
        return str(self.message)


class LazyRoot:
    """
    Allows the Root reference to be defined outside loading. (Since it cannot be defined during Loading)
    """

    __slots__ = "__root"

    def __init__(self) -> None:
        self.__root: Root = None

    def _set_root(self, root: typ.Any) -> None:
        self.__root = root

    @property
    def root(self) -> Root:
        """
        Fetch the Root.
        """
        return self.__root

    @staticmethod
    def with_root(root: tabc.Mapping | Root) -> "LazyRoot":
        lazy_root = LazyRoot()
        lazy_root._set_root(root)
        return lazy_root


class LazyEval(abc.ABC, typ.Generic[RT]):
    """
    Base class for handling the output of a Tag that needs to be run just-in-time.
    """

    tag: typ.Final[Tag]
    """
    Tag that created this instance
    """

    def __init__(self, tag: Tag) -> None:
        self.tag = tag
        self.__done = False
        self.__lock = RLock()

    @abc.abstractmethod
    def _run(self) -> RT:
        """
        Run the Tag Logic.

        :return: Result of the lazy evaluation
        :rtype: RT
        :Note: Caching does not occur within this method
        """
        ...

    @cached_property
    def __result(self) -> RT:
        return self._run()

    def __run(self) -> RT:
        if self.__done:
            return self.__result
        else:
            with self.__lock:
                result = self.__result
                self.__done = True

            del self.__lock
            return result

    @cached_property
    def result(self) -> RT | typ.Any:
        """
        Result of the lazy evaluation, completing any chains. (Cached)
        """
        result = self.__run()
        while isinstance(result, LazyEval):
            result = result.__run()
        return result

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.tag}>"

    def __deepcopy__(self, memo: dict[int, typ.Any]) -> LazyEval:
        # Don't copy LazyEval's
        return self

    def __copy__(self) -> LazyEval:
        # Don't copy LazyEval's
        return self


@dataclass(frozen=True, kw_only=True, slots=True)
class LoadOptions:
    """
    Type: frozen :py:func:`dataclass <dataclasses.dataclass>`

    Holds the parameters used when loading the configuration file.
    """

    obj_pairs_func: typ.Type[tabc.Mapping]
    """
    Type being used for YAML mappings
    """
    sequence_func: typ.Type[tabc.Sequence]
    """
    Type being used for YAML sequences
    """
    mutable: bool
    """
    Value of the mutable flag
    """
    file_location: Path | None
    """
    Path of the file being loaded
    """
    relative_to_directory: Path
    """
    Path for making relative file paths
    """
    previous: LoadOptions | None
    """
    Pointer to previous options, if this file was loaded by another
    """


@dataclass(frozen=True, kw_only=True, slots=True)
class StateHolder:
    """
    Type: frozen :py:func:`dataclass <dataclasses.dataclass>`

    Used to pass state define while Loading configuration files into Tags.
    """

    options: LoadOptions
    """
    Options from Loading
    """
    lazy_root_obj: LazyRoot
    """
    Shared reference to the final root configuration
    """



from __future__ import annotations

import collections.abc as tabc
import copy
import json
import sys
import typing as typ
from itertools import starmap
from weakref import ReferenceType, ref

from granular_configuration_language._base_path import BasePathPart
from granular_configuration_language._s import setter_secret
from granular_configuration_language.exceptions import InvalidBasePathException, PlaceholderConfigurationError
from granular_configuration_language.yaml.classes import RT, LazyEval, P, Placeholder, T

if sys.version_info >= (3, 11):
    from typing import Generic, TypedDict, Unpack, dataclass_transform
elif typ.TYPE_CHECKING:
    from typing_extensions import Generic, TypedDict, Unpack, dataclass_transform
else:

    def dataclass_transform(**kwargs: typ.Any) -> typ.Callable[[typ.Callable[P, RT]], typ.Callable[P, RT]]:
        def identity(func: typ.Callable[P, RT]) -> typ.Callable[P, RT]:
            return func

        return identity


if typ.TYPE_CHECKING:

    class Kwords_typed_get(Generic[T], TypedDict, total=False):
        default: T
        predicate: tabc.Callable[[typ.Any], typ.TypeGuard[T]]


class AttributeName(tabc.Iterable[str]):
    __slots__ = ("__prev", "__explicit_prev", "__name", "__weakref__")

    def __init__(
        self,
        name: typ.Any,
        *,
        prev: ReferenceType[AttributeName] | None = None,
        explicit_prev: tabc.Iterable[str] = tuple(),
    ) -> None:
        self.__prev = prev
        self.__explicit_prev = explicit_prev
        self.__name = name

    @staticmethod
    def as_root() -> AttributeName:
        return AttributeName("$", explicit_prev=tuple())

    def append_suffix(self, name: typ.Any) -> AttributeName:
        return AttributeName(name, prev=ref(self))

    def with_suffix(self, name: typ.Any) -> str:
        return ".".join(self._plus_one(name))

    def __iter__(self) -> tabc.Iterator[str]:
        if self.__prev:
            yield from self.__prev() or tuple()
        else:
            yield from self.__explicit_prev
        yield self.__name if isinstance(self.__name, str) else f"`{repr(self.__name)}`"

    def _plus_one(self, last: str) -> tabc.Iterator[str]:
        yield from self
        yield last if isinstance(last, str) else f"`{repr(last)}`"

    def __str__(self) -> str:
        return ".".join(self)


@dataclass_transform(frozen_default=True, eq_default=True, kw_only_default=True)
class Configuration(tabc.Mapping[typ.Any, typ.Any]):
    r"""
    This class represents an immutable :py:class:`~collections.abc.Mapping` of
    configuration.

    You can create type annotated subclasses of :py:class:`Configuration` to
    enable type checking and code completion, as if your subclass was a
    :py:func:`dataclass <dataclasses.dataclass>` [#f1]_.

    With you typed class, you can cast a general :py:class:`.Configuration` to
    your subclass via :py:meth:`Configuration.as_typed`.

    .. code-block:: python

        class SubConfig(Configuration):
            c: str

        class Config(Configuration):
            a: int
            b: SubConfig


        config = ... # A Configuration instance
        typed = config.as_typed(Config)

        assert typed.a == 101
        assert typed.b.c == "test me"
        assert typed["a"] == 101

        # Or loading with LazyLoadConfiguration

        typed = LazyLoadConfiguration("config.yaml").as_typed(Config)

    .. note::

        You should use :py:meth:`LazyLoadConfiguration.as_typed` to load as a
        typed :py:class:`.Configuration`.

    .. [#f1]

        See :py:func:`~typing.dataclass_transform` → "on base class" for
        implementation details
    """

    @typ.overload
    def __init__(self) -> None: ...

    @typ.overload
    def __init__(self, mapping: tabc.Mapping[typ.Any, typ.Any]) -> None: ...

    @typ.overload
    def __init__(self, iterable: tabc.Iterable[tuple[typ.Any, typ.Any]]) -> None: ...

    @typ.overload
    def __init__(self, **kwargs: typ.Any) -> None: ...

    def __init__(self, *arg: tabc.Mapping | tabc.Iterable[tuple[typ.Any, typ.Any]], **kwargs: typ.Any) -> None:
        self.__data: dict[typ.Any, typ.Any] = dict(*arg, **kwargs)
        self.__attribute_name = AttributeName.as_root()

    #################################################################
    # Required for Mapping
    #################################################################

    def __iter__(self) -> tabc.Iterator:
        return iter(self.__data)

    def __len__(self) -> int:
        return len(self.__data)

    def __getitem__(self, name: typ.Any) -> typ.Any:
        try:
            value = self.__data[name]
        except KeyError:
            # Start the stack trace here
            if isinstance(name, BasePathPart):
                raise InvalidBasePathException(
                    f"Base Path `{self.__attribute_name.with_suffix(name)}` does not exist."
                ) from None
            else:
                raise KeyError(repr(name)) from None

        if isinstance(value, LazyEval):
            try:
                value = value.result
                self._private_set(name, value, setter_secret)
            except RecursionError as e:
                raise RecursionError(
                    f"{value.tag} at `{self.__attribute_name.with_suffix(name)}` caused a recursion error: {e}"
                ) from None

        if isinstance(value, Placeholder):
            raise PlaceholderConfigurationError(
                f'!Placeholder at `{self.__attribute_name.with_suffix(name)}` was not overwritten. Message: "{value}"'
            )

        if isinstance(value, Configuration):
            value.__attribute_name = self.__attribute_name.append_suffix(name)
            return value
        else:
            return value

    #################################################################
    # Overridden Mapping methods
    #################################################################

    def __contains__(self, key: typ.Any) -> bool:
        return key in self.__data

    def get(self, key: typ.Any, default: typ.Any = None) -> typ.Any:
        """
        Return the value for key if key is in the :py:class:`Configuration`, else default.

        Parameters:
            key (~typing.Any): Key being fetched
            default (~typing.Any, optional): Default value. Defaults to :py:data:`None`.

        Returns:
            ~typing.Any: Value
        """
        return self[key] if self.exists(key) else default

    #################################################################
    # Required behavior overrides
    #################################################################

    def __repr__(self) -> str:
        return repr(self.__data)

    def __deepcopy__(self, memo: dict[int, typ.Any]) -> Configuration:
        other = Configuration()
        memo[id(self)] = other
        other.__data = copy.deepcopy(self.__data, memo=memo)
        return other

    def __copy__(self) -> Configuration:
        other = Configuration()
        other.__data = copy.copy(self.__data)
        return other

    copy = __copy__

    #################################################################
    # Internal methods
    #################################################################

    def _private_set(self, key: typ.Any, value: typ.Any, secret: object) -> None:
        if secret is setter_secret:
            self.__data[key] = value
        else:
            raise TypeError("`_private_set` is private and not for external use")

    def _raw_items(self) -> tabc.Iterator[tuple[typ.Any, typ.Any]]:
        return map(lambda key: (key, self.__data[key]), self)

    #################################################################
    # Public interface methods
    #################################################################

    def __getattr__(self, name: str) -> typ.Any:
        """
        Provides a potentially cleaner path as an alternative to :py:meth:`~object.__getitem__`.

        Throws :py:exc:`AttributeError` instead of :py:exc:`KeyError`, as compared to :py:meth:`~object.__getitem__`
        when an attribute is not present.

        :example:
            .. code-block:: python

                config.a.b.c          # Using `__getattr__`
                config["a"]["b"]["c"] # Using `__getitem__`
        :param str name: Attribute name
        :return: Fetched value
        :rtype:  ~typing.Any
        :raises AttributeError: When an attribute is not present.
        """
        if name not in self:
            raise AttributeError(f"Request attribute `{self.__attribute_name.with_suffix(name)}` does not exist")

        return self[name]

    def exists(self, key: typ.Any) -> bool:
        """
        Checks that a key exists and is not a :py:class:`~.Placeholder`

        Parameters:
            key (~typing.Any): key to be checked

        Returns:
            bool: Returns :py:data:`True` if the key exists and is not a :py:class:`~.Placeholder`
        """
        return (key in self) and not isinstance(self.__data[key], Placeholder)

    def evaluate_all(self) -> None:
        """
        Evaluates all lazy tag functions and throws an exception on :py:class:`~.Placeholder` instances
        """

        for value in self.values():
            if isinstance(value, Configuration):
                value.evaluate_all()

    def as_dict(self) -> dict[typ.Any, typ.Any]:
        """
        Returns this :py:class:`Configuration` as standard Python :py:class:`dict`.
        Nested :class:`Configuration` objects will also be converted.

        :return: A shallow :py:class:`dict` copy
        :rtype: dict
        :note: This will evaluate all lazy tag functions and throw an exception on :py:class:`~.Placeholder` objects.
        """
        return dict(
            starmap(
                lambda key, value: (key, value.as_dict() if isinstance(value, Configuration) else value),
                self.items(),
            )
        )

    def as_json_string(self, *, default: tabc.Callable[[typ.Any], typ.Any] | None = None, **kwds: typ.Any) -> str:
        r"""
        Returns this :py:class:`Configuration` as a JSON string, using standard :py:mod:`json`
        library and (as default) the default factory provided by this library
        (:py:func:`granular_configuration_language.json_default`).

        :param \~typing.Callable[[\~typing.Any], \~typing.Any], optional default: Replacement ``default`` factory. Defaults to :py:func:`~granular_configuration_language.json_default`.
        :param ~typing.Any \*\*kwds: Arguments to be passed into :py:func:`json.dumps`
        :return: JSON-format string
        :rtype: str
        :note: This will evaluate all lazy tag functions and throw an exception on :py:class:`~.Placeholder` objects.
        """
        from granular_configuration_language import json_default

        return json.dumps(self, default=default or json_default, **kwds)

    @typ.overload
    def typed_get(self, type: typ.Type[T], key: typ.Any) -> T: ...

    @typ.overload
    def typed_get(self, type: typ.Type[T], key: typ.Any, *, default: T) -> T: ...

    @typ.overload
    def typed_get(
        self, type: typ.Type[T], key: typ.Any, *, predicate: tabc.Callable[[typ.Any], typ.TypeGuard[T]]
    ) -> T: ...

    @typ.overload
    def typed_get(
        self, type: typ.Type[T], key: typ.Any, *, default: T, predicate: tabc.Callable[[typ.Any], typ.TypeGuard[T]]
    ) -> T: ...

    def typed_get(self, type: typ.Type[T], key: typ.Any, **kwds: Unpack[Kwords_typed_get[T]]) -> T:
        r"""
        Provides a typed-checked :py:meth:`get` option

        Parameters:
            type (type[T]): Wanted typed
            key (~typing.Any): Key for wanted value
            default (T, optional): Provides a default value like :py:meth:`dict.get`
            predicate (\~typing.Callable[[~typing.Any], ~typing.TypeGuard[T]], optional):
                Replaces the ``isinstance(value, type)`` check with a custom
                method ``predicate(value: Any) -> bool``

        Returns:
            T: Value stored under the key

        Raises:
            TypeError: If the real type is not an instance of the expected type
        """

        try:
            value = self[key]
        except KeyError:
            if "default" in kwds:
                return kwds["default"]
            else:
                raise

        if (("predicate" in kwds) and kwds["predicate"](value)) or isinstance(value, type):
            return value
        else:
            raise TypeError(f"Incorrect type. Got: `{repr(value)}`. Wanted: `{repr(type)}`")

    def as_typed(self, typed_base: type[C]) -> C:
        """
        Cast this :py:class:`Configuration` instance into subclass of :py:class:`Configuration` with typed annotated attributes

        :param type[C] typed_base: Subclass of :py:class:`Configuration` to assume
        :return: This instance
        :rtype: C
        :note: No runtime typing check occurs.
        """
        return typ.cast(C, self)


class MutableConfiguration(tabc.MutableMapping[typ.Any, typ.Any], Configuration):
    """
    This class represents an :py:class:`~collections.abc.MutableMapping` of the configuration. Inherits from :py:class:`Configuration`
    """

    if typ.TYPE_CHECKING:
        # For Pylance. Until 3.10 is removed

        def __init__(self, *arg: tabc.Mapping | tabc.Iterable[tuple[typ.Any, typ.Any]], **kwargs: typ.Any) -> None:
            super().__init__(*arg, **kwargs)

    # Remember `Configuration.__data` is really `Configuration._Configuration__data`
    # Type checkers do ignore this fact, because this is something to be avoided.
    # I want to continue to use self.__data to avoid people being tempted to reach in.

    def __delitem__(self, key: typ.Any) -> None:
        del self._Configuration__data[key]

    def __setitem__(self, key: typ.Any, value: typ.Any) -> None:
        self._Configuration__data[key] = value

    def __deepcopy__(self, memo: dict[int, typ.Any]) -> MutableConfiguration:
        other = MutableConfiguration()
        memo[id(self)] = other
        # Use setattr to avoid mypy and pylance being confused
        setattr(other, "_Configuration__data", copy.deepcopy(self._Configuration__data, memo=memo))
        return other

    def __copy__(self) -> MutableConfiguration:
        other = MutableConfiguration()
        # Use setattr to avoid mypy and pylance being confused
        setattr(other, "_Configuration__data", copy.copy(self._Configuration__data))
        return other

    copy = __copy__


C = typ.TypeVar("C", bound=Configuration)



from __future__ import annotations

import collections.abc as tabc
import importlib
import inspect
import os
import sys

from granular_configuration_language.exceptions import DoesNotExist, IsNotAClass, IsNotCallable
from granular_configuration_language.yaml.decorators import Tag, as_lazy, interpolate_value_without_ref, string_tag


def add_cwd_to_path() -> None:
    cwd = os.getcwd()
    if sys.path[0] != cwd:  # pragma: no cover
        sys.path.insert(0, cwd)


def get_func(func_path: str) -> tabc.Callable:
    add_cwd_to_path()
    mod_name, func_name = func_path.rsplit(".", 1)
    try:
        func: tabc.Callable = getattr(importlib.import_module(mod_name), func_name)
        return func
    except (ImportError, AttributeError):
        raise DoesNotExist(f"Could not load {func_path}")


@string_tag(Tag("!Class"))
@as_lazy
@interpolate_value_without_ref
def class_handler(value: str) -> tabc.Callable:
    class_type = get_func(value)
    if inspect.isclass(class_type):
        return class_type
    else:
        raise IsNotAClass(f"Classes loaded by !Class must pass `inspect.isclass`: `{value}` is not a class")


@string_tag(Tag("!Func"))
@as_lazy
@interpolate_value_without_ref
def func_handler(value: str) -> tabc.Callable:
    func = get_func(value)
    if callable(func):
        return func
    else:
        raise IsNotCallable(f"Functions loaded by !Func must be callable: `{value}` is not callable")



from __future__ import annotations

import collections.abc as tabc
import typing as typ

from granular_configuration_language._configuration import Configuration
from granular_configuration_language.yaml.decorators import (
    LoadOptions,
    Tag,
    as_lazy_with_load_options,
    sequence_of_any_tag,
)


@sequence_of_any_tag(Tag("!Merge"))
@as_lazy_with_load_options
def handler(value: tabc.Sequence[typ.Any], options: LoadOptions) -> Configuration:
    from granular_configuration_language import merge

    return merge(value, mutable=options.mutable)



from __future__ import annotations

import typing as typ
from pathlib import Path

from granular_configuration_language._configuration import Configuration, MutableConfiguration
from granular_configuration_language.yaml.classes import LazyEval, LazyRoot, LoadOptions, StateHolder
from granular_configuration_language.yaml.load._handler import load_yaml_string


def loads(
    config_str: str,
    *,
    lazy_root: LazyRoot | None = None,
    file_path: Path | None = None,
    previous_options: LoadOptions | None = None,
    mutable: bool = False,
) -> typ.Any:
    state = StateHolder(
        lazy_root_obj=lazy_root or LazyRoot(),
        options=LoadOptions(
            file_location=file_path,
            relative_to_directory=file_path.parent if file_path is not None else Path(),
            obj_pairs_func=obj_pairs_func(mutable),
            sequence_func=sequence_func(mutable),
            mutable=mutable,
            previous=previous_options,
        ),
    )

    result = load_yaml_string(config_str, state)

    if lazy_root is None:
        state.lazy_root_obj._set_root(result)

    if isinstance(result, LazyEval):
        return result.result
    else:
        return result


def obj_pairs_func(mutable: bool) -> typ.Type[Configuration] | typ.Type[MutableConfiguration]:
    return MutableConfiguration if mutable else Configuration


def sequence_func(mutable: bool) -> typ.Type[list] | typ.Type[tuple]:
    return list if mutable else tuple



from __future__ import annotations

from granular_configuration_language.yaml.decorators import (
    Root,
    Tag,
    as_lazy_with_root,
    interpolate_value_with_ref,
    string_tag,
)
from granular_configuration_language.yaml.decorators.ref import resolve_json_ref


@string_tag(Tag("!Ref"))
@as_lazy_with_root
@interpolate_value_with_ref
def handler(value: str, root: Root) -> str:
    return resolve_json_ref(value, root)



from __future__ import annotations

from granular_configuration_language.yaml.classes import LoadOptions, Root
from granular_configuration_language.yaml.decorators._base import Tag, TagDecoratorBase
from granular_configuration_language.yaml.decorators._lazy import (
    as_lazy,
    as_lazy_with_load_options,
    as_lazy_with_root,
    as_lazy_with_root_and_load_options,
    as_not_lazy,
)
from granular_configuration_language.yaml.decorators._type_checking import (
    mapping_of_any_tag,
    sequence_of_any_tag,
    string_or_twople_tag,
    string_tag,
)
from granular_configuration_language.yaml.decorators.interpolate import (
    interpolate_value_with_ref,
    interpolate_value_without_ref,
)



from __future__ import annotations

import collections.abc as tabc
import sys
import typing as typ
from pathlib import Path

from granular_configuration_language.exceptions import ParsingTriedToCreateALoop
from granular_configuration_language.yaml.classes import LoadOptions

if sys.version_info >= (3, 12):

    def walkup(file: Path, relative_to: Path) -> Path:
        return file.relative_to(relative_to, walk_up=True)

else:
    import os

    def _get_segments(path: os.PathLike) -> tabc.Iterator[str]:
        head, tail = os.path.split(path)
        if tail:
            yield from _get_segments(head)
            yield tail
        else:
            yield head

    def get_segments(path: os.PathLike) -> list[str]:
        return list(_get_segments(path))

    def walkup(file: Path, relative_to: Path) -> Path:
        # Modified from the 3.12 pathlib.PurePath.relative_to implementation

        for step, path in enumerate([relative_to] + list(relative_to.parents)):
            if file.is_relative_to(path):
                break
            elif path.name == "..":
                raise ValueError(f"'..' segment in {str(relative_to)!r} cannot be walked")
        else:
            raise ValueError(f"{str(file)!r} and {str(relative_to)!r} have different anchors")
        parts = [".."] * step + get_segments(file)[len(get_segments(path)) :]
        return Path(*parts)


FILE_EXTENSION: typ.Final = ".environment_variable-a5b55071-b86e-4f22-90fc-c9db335691f6"


def _get_chain_reversed(options: LoadOptions) -> tabc.Iterator[str]:
    relative_to = Path().resolve()
    seen: set[str] = set()

    if options.previous:
        yield from _get_chain_reversed(options.previous)

    if not options.file_location:
        pass
    elif options.file_location.suffix == FILE_EXTENSION:
        yield "$" + options.file_location.stem
    elif options.file_location.name not in seen:
        seen.add(options.file_location.name)
        yield options.file_location.name
    else:
        try:
            yield str(walkup(options.file_location, relative_to))
        except ValueError:
            yield "?/" + options.file_location.name


def is_in_chain(file: Path, options: LoadOptions) -> bool:
    # Note *.environment_variable don't exist, so .resolve() and .samefile() fail

    if (
        options.file_location
        and (file.name == options.file_location.name)
        and (file == options.file_location or file.samefile(options.file_location))
    ):
        return True
    elif options.previous:
        return is_in_chain(file, options.previous)
    else:
        return False


def make_chain_message(tag: str, value: str, options: LoadOptions) -> ParsingTriedToCreateALoop:
    return ParsingTriedToCreateALoop(
        f"`{tag} {value}` tried to load itself in chain: ({'→'.join(_get_chain_reversed(options))}→...)"
    )


def create_environment_variable_path(env_var: str) -> Path:
    return Path(env_var + FILE_EXTENSION)



from __future__ import annotations

from granular_configuration_language.yaml.decorators.interpolate._interpolate import (
    interpolate_value_with_ref,
    interpolate_value_without_ref,
    interpolation_needs_ref_condition,
)



from __future__ import annotations

import abc
import collections.abc as tabc
import dataclasses
import typing as typ
from functools import wraps

from ruamel.yaml import MappingNode, Node, SafeConstructor, ScalarNode, SequenceNode

from granular_configuration_language.exceptions import ErrorWhileLoadingTags
from granular_configuration_language.yaml.classes import RT, StateHolder, T, Tag
from granular_configuration_language.yaml.load._constructors import construct_mapping, construct_sequence


@dataclasses.dataclass(frozen=True, eq=False, slots=True, repr=False)
class TagConstructor:
    """
    Links the YAML Constructor to the Tag Logic
    """

    tag: Tag
    friendly_type: str
    constructor: tabc.Callable[[typ.Type[SafeConstructor], StateHolder], None]

    def __eq__(self, value: object) -> bool:
        return (isinstance(value, self.__class__) and self.tag == value.tag) or (
            isinstance(value, str) and self.tag == value
        )

    def __hash__(self) -> int:
        return hash(self.tag)

    def __call__(self, constructor: typ.Type[SafeConstructor], state: StateHolder) -> None:
        return self.constructor(constructor, state)

    def __repr__(self) -> str:
        return f"<TagConstructor(`{self.tag}`): {self.constructor.__module__}.{self.constructor.__name__}>"


class TagDecoratorBase(typ.Generic[T], abc.ABC):
    """Base class for Tag Decorator factories.

    You must implement the :py:attr:`user_friendly_type` property and define the generic type.

    Example:

    .. code-block:: python

        class string_tag(TagDecoratorBase[str]):
            Type: typing.TypeAlias = str

            @property
            def user_friendly_type(self) -> str:
                return "str"

    You must override at least one of :py:meth:`scalar_node_type_check`, :py:meth:`sequence_node_type_check`,or
    :py:meth:`mapping_node_type_check`.

    - For :py:meth:`scalar_node_type_check` to be called the YAML has already be tested to be a :py:class:`str`.
    - For :py:meth:`sequence_node_type_check` to be called the YAML has already be tested to be a
      :py:class:`~collections.abc.Sequence`.
    - For :py:meth:`mapping_node_type_check` to be called the YAML has already be tested to be a
      :py:class:`~collections.abc.Mapping`.

    If these are enough, then you may just return :py:data:`True` in the override method. Otherwise, implement the
    override as a :py:data:`~typing.TypeGuard`.

    If the value needs to be altered before being passed to Tag functions, override :py:meth:`scalar_node_transformer`,
    :py:meth:`scalar_node_transformer`, or :py:meth:`mapping_node_transformer`, as needed.

    The transformer is called if the associated node type check passes, just before the value is passed to tag function.

    """

    __slots__ = ("tag",)

    def __init__(self, tag: Tag) -> None:
        self.tag: typ.Final = tag

        if not tag.startswith("!"):
            raise ErrorWhileLoadingTags(f"Tag `{tag}` error: All tags must begin with `!`.")

    @property
    @abc.abstractmethod
    def user_friendly_type(self) -> str:
        """User-friendly of the type expected by Tag Decorator.

        Note:
            - Use Python types for consistent communication.
            - This is used when generating exception messages.

        :returns: User-friendly string
        :rtype: str

        """
        ...

    def scalar_node_type_check(self, value: str) -> typ.TypeGuard[T]:
        """Defaults to :py:data:`False`. Override to enable Scalar Node support.

        Parameters:
            value (str): YAML value

        Returns:
            ~typing.TypeGuard[T]: Return :py:data:`True`, if ``value`` is supported.

        """
        return False

    def sequence_node_type_check(self, value: tabc.Sequence) -> typ.TypeGuard[T]:
        """Defaults to :py:data:`False`. Override to enable Sequence Node support.

        Parameters:
            value (~collections.abc.Sequence): YAML value

        Returns:
            ~typing.TypeGuard[T]: Return :py:data:`True`, if ``value`` is supported.

        """
        return False

    def mapping_node_type_check(self, value: tabc.Mapping) -> typ.TypeGuard[T]:
        """Defaults to :py:data:`False`. Override to enable Mapping Node support.

        Parameters:
            value (~collections.abc.Mapping): YAML value

        Returns:
            ~typing.TypeGuard[T]: Return :py:data:`True`, if ``value`` is supported.

        """
        return False

    def scalar_node_transformer(self, value: typ.Any) -> T:
        """Defaults to an identity operation. Override if the value needs to be altered before being passed to Tag functions.

        Only called if :py:meth:`scalar_node_type_check` return :py:data:`True`.

        As an example, a float tag could be supported by:

        .. code-block:: python

            def scalar_node_type_check(self, value: Any) -> TypeGuard[float]:
                try:
                    float(value)
                    return True
                except ValueError:
                    return False


            def scalar_node_transformer(self, value: Any) -> float:
                return float(value)

        Parameters:
            value (~typing.Any): YAML value

        Returns:
            T: Transformed value

        """
        return value

    def sequence_node_transformer(self, value: typ.Any) -> T:
        """Defaults to an identity operation. Override if the value needs to be altered before being passed to Tag functions.

        Only called if :py:meth:`sequence_node_type_check` return :py:data:`True`.

        Parameters:
            value (~typing.Any): YAML value

        Returns:
            T: Transformed value

        """
        return value

    def mapping_node_transformer(self, value: typ.Any) -> T:
        """Defaults to an identity operation. Override if the value needs to be altered before being passed to Tag functions.

        Only called if :py:meth:`mapping_node_type_check` return :py:data:`True`.

        Parameters:
            value (~typing.Any): YAML value

        Returns:
            T: Transformed value

        """
        return value

    def __call__(self, handler: tabc.Callable[[Tag, T, StateHolder], RT]) -> TagConstructor:
        # """Takes the wrapped tag function and further wraps it for configuration loading.
        # :param (~collections.abc.Callable[[Tag, T, StateHolder], RT]) handler: Wrapped Tag Function
        # :return: Tag Function ready to be used when loading configuration
        # :rtype: TagConstructor
        # :meta private:
        # """
        # autodoc refuses to exclude `__call__`, even though this is not a part of the public interface

        # Don't capture self in the function generation
        tag = self.tag
        user_friendly_type = self.user_friendly_type
        scalar_node_type_check = self.scalar_node_type_check
        sequence_node_type_check = self.sequence_node_type_check
        mapping_node_type_check = self.mapping_node_type_check
        scalar_node_transformer = self.scalar_node_transformer
        sequence_node_transformers = self.sequence_node_transformer
        mapping_node_transformer = self.mapping_node_transformer

        @wraps(handler)
        def add_handler(
            constructor: typ.Type[SafeConstructor],
            state: StateHolder,
        ) -> None:
            @wraps(handler)
            def type_handler(constructor: SafeConstructor, node: Node) -> RT:
                if isinstance(node, ScalarNode):
                    value = constructor.construct_scalar(node)
                    if isinstance(value, str) and scalar_node_type_check(value):
                        return handler(tag, scalar_node_transformer(value), state)
                elif isinstance(node, SequenceNode):
                    value = construct_sequence(state.options.sequence_func, constructor, node)
                    if isinstance(value, tabc.Sequence) and sequence_node_type_check(value):
                        return handler(tag, sequence_node_transformers(value), state)
                elif isinstance(node, MappingNode):
                    value = construct_mapping(state.options.obj_pairs_func, constructor, node)
                    if isinstance(value, tabc.Mapping) and mapping_node_type_check(value):
                        return handler(tag, mapping_node_transformer(value), state)
                else:
                    pass  # pragma: no cover

                # Fallback Exception
                raise ValueError(f"`{tag}` supports: {user_friendly_type}. Got: `{repr(node)}`")

            constructor.add_constructor(tag, type_handler)

        return TagConstructor(tag, user_friendly_type, add_handler)



from __future__ import annotations

import collections.abc as tabc
from operator import attrgetter

tracker_id = attrgetter("__module__", "__name__", "__qualname__")

with_ref: set[tuple[str, str, str]] = set()
without_ref: set[tuple[str, str, str]] = set()
not_lazy: set[tuple[str, str, str]] = set()


def track_as_with_ref(func: tabc.Callable) -> None:
    with_ref.add(tracker_id(func))


def is_with_ref(func: tabc.Callable) -> bool:
    return tracker_id(func) in with_ref


def track_as_without_ref(func: tabc.Callable) -> None:
    without_ref.add(tracker_id(func))


def is_without_ref(func: tabc.Callable) -> bool:
    return tracker_id(func) in without_ref


def track_as_not_lazy(func: tabc.Callable) -> None:
    not_lazy.add(tracker_id(func))


def is_not_lazy(func: tabc.Callable) -> bool:
    return tracker_id(func) in not_lazy



from __future__ import annotations

import sys
from datetime import date, datetime
from functools import partial

from granular_configuration_language.yaml.decorators import Tag, as_lazy, interpolate_value_without_ref, string_tag

if sys.version_info >= (3, 11):
    date_fromisoformat = date.fromisoformat
    datetime_fromisoformat = datetime.fromisoformat
else:  # if sys.version_info
    from dateutil.parser import parse

    datetime_fromisoformat = partial(parse, yearfirst=True, dayfirst=False)
    date_fromisoformat = lambda value: datetime_fromisoformat(value).date()


@string_tag(Tag("!Date"))
@as_lazy
@interpolate_value_without_ref
def date_handler(value: str) -> date:
    return date_fromisoformat(value)


@string_tag(Tag("!DateTime"))
@as_lazy
@interpolate_value_without_ref
def datetime_handler(value: str) -> date:
    dt = datetime_fromisoformat(value)
    return dt



from __future__ import annotations

from granular_configuration_language.yaml.decorators import Tag, as_not_lazy, string_tag


@string_tag(Tag("!Del"))
@as_not_lazy
def handler(value: str) -> str:
    return value



from __future__ import annotations

import collections.abc as tabc
import inspect
import json
import typing as typ
from collections.abc import Mapping, Sequence
from datetime import date, datetime
from functools import partial, update_wrapper
from uuid import UUID

from granular_configuration_language import Configuration, LazyLoadConfiguration
from granular_configuration_language.proxy import SafeConfigurationProxy


def get_name(value: tabc.Callable) -> str:
    try:
        return f"<{value.__module__}.{value.__name__}>"
    except Exception:  # pragma: no cover
        return f"<{repr(value)}>"


def json_default(value: typ.Any) -> typ.Any:
    """A factory function to be used by the :py:func:`json.dump` family of functions.

    Provides serialization for types produced by this library's Tags.

    Explicitly:

    - :py:class:`~.Configuration` as :py:class:`dict`
    - ``!UUID``/:py:class:`uuid.UUID` as hyphenated hex string
    - ``!Date``/:py:class:`datetime.date` as :py:meth:`~datetime.date.isoformat`
    - ``!DateTime``/:py:class:`datetime.datetime` as :py:meth:`~datetime.datetime.isoformat`
    - ``!Func``/:py:class:`~collections.abc.Callable` as ``f"<{func.__module__}.{func.__name__}>"``
    - ``!Class``/:py:class:`type` as ``f"<{class.__module__}.{class.__name__}>"``
    - For niceness, :py:class:`~collections.abc.Mapping` and non-:class:`str` :py:class:`~collections.abc.Sequence`
      instances are converted to :py:class:`dict` and :py:class:`tuple`

    :param ~typing.Any value: Value being converted

    :returns: :py:func:`json.dump` compatible object
    :rtype: Any

    :raises TypeError: When an incompatible is provided, as required by :py:class:`~json.JSONEncoder`

    """

    if isinstance(value, Configuration):
        return value.as_dict()
    elif isinstance(value, LazyLoadConfiguration):
        return value.config.as_dict()
    elif isinstance(value, SafeConfigurationProxy):
        return value.as_dict()
    elif isinstance(value, UUID):
        return str(value)
    elif isinstance(value, (date, datetime)):
        return value.isoformat()
    elif inspect.isclass(value):
        return get_name(value)
    elif isinstance(value, partial):
        return f"<{repr(value)}>"
    elif callable(value):
        return get_name(value)
    elif isinstance(value, Mapping):
        return dict(value)
    elif isinstance(value, Sequence) and not isinstance(value, str):
        return tuple(value)
    else:
        return json.JSONEncoder().default(value)


dumps = update_wrapper(partial(json.dumps, default=json_default), json.dumps)



from __future__ import annotations

import collections.abc as tabc
import copy
import os
import typing as typ
from collections.abc import Mapping, MutableMapping
from functools import cached_property
from itertools import chain

from granular_configuration_language._cache import NoteOfIntentToRead, prepare_to_load_configuration
from granular_configuration_language._configuration import C, Configuration, MutableConfiguration
from granular_configuration_language._locations import Locations, PathOrStr
from granular_configuration_language.exceptions import ErrorWhileLoadingConfig


def _read_locations(
    load_order_location: tabc.Iterable[PathOrStr],
    use_env_location: bool,
    env_location_var_name: str,
) -> Locations:
    if (use_env_location or (env_location_var_name != "G_CONFIG_LOCATION")) and (env_location_var_name in os.environ):
        env_locs = os.environ[env_location_var_name].split(",")
        load_order_location = chain(load_order_location, env_locs)
    return Locations(load_order_location)


class SafeConfigurationProxy(Mapping):
    """
    Wraps a :py:class:`.LazyLoadConfiguration` instance to proxy all method and
    attribute calls to its :py:class:`.Configuration` instance.

    :param LazyLoadConfiguration llc:
        :py:class:`.LazyLoadConfiguration` instance to be wrapped
    :note:
        Wrapping :py:class:`.LazyLoadConfiguration` maintains all laziness build
        into :py:class:`.LazyLoadConfiguration`.
    """

    __slots__ = ("__llc",)

    def __init__(self, llc: LazyLoadConfiguration) -> None:
        self.__llc = llc

    def __getattr__(self, name: str) -> typ.Any:
        return getattr(self.__llc.config, name)

    def __getitem__(self, key: typ.Any) -> typ.Any:
        return self.__llc.config[key]

    def __iter__(self) -> tabc.Iterator[typ.Any]:
        return iter(self.__llc.config)

    def __len__(self) -> int:
        return len(self.__llc.config)

    def __contains__(self, key: typ.Any) -> bool:
        return key in self.__llc.config

    def __deepcopy__(self, memo: dict[int, typ.Any]) -> Configuration:
        return copy.deepcopy(self.__llc.config, memo=memo)

    def __copy__(self) -> Configuration:
        return copy.copy(self.__llc.config)

    copy = __copy__

    def __repr__(self) -> str:
        return repr(self.__llc.config)


class LazyLoadConfiguration(Mapping):
    r"""
    Provides a lazy interface for loading Configuration from file paths on first access.

    You can optionally enable pull locations from an environment variable.

    See :py:meth:`LazyLoadConfiguration.as_typed` for type annotated usage.

    :param ~pathlib.Path | str | os.PathLike load_order_location:
            File path to configuration file
    :param str | ~collections.abc.Sequence[str], optional base_path:
        Defines the subsection of the configuration file to use. See Examples for usage options.
    :param bool, optional use_env_location:
        Enabled to use the default environment variable location.
    :param str, optional env_location_var_name:
        Specify what environment variable to check for additional file paths.
    :param Configuration, optional inject_before:
        Inject a runtime :py:class:`.Configuration` instance as if it were the first load file.
    :param Configuration, optional inject_after:
        Inject a runtime :py:class:`.Configuration` instance as if it were the last load file.
    :param bool, optional disable_caching:
        When :py:data:`True`, caching of "identical immutable configurations" is disabled.

    :examples:
        .. code-block:: python

            # Base Path Examples
            LazyLoadConfiguration(..., base_path="base_path")  # Single Key
            LazyLoadConfiguration(..., base_path="/base/path")  # JSON Pointer (strings only)
            LazyLoadConfiguration(..., base_path=("base", "path"))  # List of keys

            # Use Environment Variable: "CONFIG_LOC"
            LazyLoadConfiguration(..., env_location_var_name="CONFIG_LOC")

            # Use default Environment Variable: "G_CONFIG_LOCATION"
            LazyLoadConfiguration(..., use_env_location=True)

            # With a typed `Configuration`
            LazyLoadConfiguration( ... ).as_typed(TypedConfig)

    :note:
        - The Environment Variable is read as a comma-delimited list of configuration path that will be appended to ``load_order_location`` list.
        - Setting the Environment Variable is always optional.
        - Setting ``use_env_location`` to :py:data:`True` is only required if you don't change ``env_location_var_name`` from its default value of ``G_CONFIG_LOCATION``.
        - ``inject_before`` and ``inject_after`` allow you to introduce dynamic settings into you configuration.

          - These injections must use :py:class:`.Configuration` for all mappings. Otherwise, they will be treated as a normal value and not merged.
          - This is only available for :py:class:`.LazyLoadConfiguration`, as :py:class:`.MutableLazyLoadConfiguration` doesn't required this.
          - Examples:

            - You might want to have a setting the is the current date. ``CONFIG.today``
            - You want to provide substitution options via ``!Sub``.  ``!Sub ${$.VALUE_DYNAMICALLY_SET_BY_LIBRARY}``
    """

    def __init__(
        self,
        *load_order_location: PathOrStr,
        base_path: str | tabc.Sequence[str] | None = None,
        use_env_location: bool = False,
        env_location_var_name: str = "G_CONFIG_LOCATION",
        inject_before: Configuration | None = None,
        inject_after: Configuration | None = None,
        disable_caching: bool = False,
        **kwargs: typ.Any,
    ) -> None:

        self.__receipt: NoteOfIntentToRead | None = prepare_to_load_configuration(
            locations=_read_locations(load_order_location, use_env_location, env_location_var_name),
            base_path=base_path,
            mutable_configuration=kwargs.get("_mutable_configuration", False),
            inject_before=inject_before,
            inject_after=inject_after,
            disable_cache=disable_caching,
        )

    def __getattr__(self, name: str) -> typ.Any:
        """Loads (if not loaded) and fetches from the underlying `Configuration` object

        :param str name: Attribute name

        :returns: Result
        :rtype: ~typing.Any

        :note: This also exposes the methods of :py:class:`Configuration` (except dunders).

        """
        return getattr(self.config, name)

    @property
    def config(self) -> Configuration:
        """Load and fetch the configuration. Configuration is cached for subsequent calls.

        :note: Loading the configuration is thread-safe and locks while the configuration is loaded to prevent
            duplicative processing and data

        """
        config = self.__config
        self.__receipt = None  # self.__config is cached
        return config

    @cached_property
    def __config(self) -> Configuration:
        if self.__receipt:
            return self.__receipt.config
        else:
            raise ErrorWhileLoadingConfig(
                "Config reference was lost before `cached_property` cached it."
            )  # pragma: no cover

    def load_configuration(self) -> None:
        """Force load the configuration."""
        # load_configuration existed prior to config, being a cached_property.
        # Now that logic is in the cached_property, so this legacy/clear code just calls the property
        self.config

    def __getitem__(self, key: typ.Any) -> typ.Any:
        return self.config[key]

    def __iter__(self) -> tabc.Iterator[typ.Any]:
        return iter(self.config)

    def __len__(self) -> int:
        return len(self.config)

    def as_typed(self, typed_base: typ.Type[C]) -> C:
        """
        Create a proxy that is cast to provide :py:class:`Configuration` subclass with typed annotated attributes.

        This proxy ensures laziness is preserved and is fully compatible with :py:class:`Configuration`.

        .. admonition:: Use as
            :class: hint

            .. code-block:: python

                    CONFIG = LazyLoadConfiguration("config.yaml").as_typed(Config)


        :param type[C] typed_base: Subclass of :py:class:`Configuration` to assume
        :return: :py:class:`.SafeConfigurationProxy` instance that has been cast to the provided type.
        :rtype: C
        :note: No runtime typing check occurs.
        """
        return typ.cast(C, SafeConfigurationProxy(self))


class MutableLazyLoadConfiguration(LazyLoadConfiguration, MutableMapping):
    r"""
    Provides a lazy interface for loading Configuration from file paths on first access.

    :py:class:`.MutableLazyLoadConfiguration` uses: :py:class:`.MutableConfiguration` for mappings and :py:class:`list` for sequences.

    You can optionally enable pull locations from an environment variable.

    :param ~pathlib.Path | str | os.PathLike load_order_location:
            File path to configuration file
    :param str | ~collections.abc.Sequence[str], optional base_path:
        Defines the subsection of the configuration file to use. See Examples for usage options.
    :param bool, optional use_env_location:
        Enabled to use the default environment variable location. (O)
    :param str, optional env_location_var_name:
        Specify what environment variable to check for additional file paths.

    :examples:
        .. code-block:: python

            # Base Path Examples
            MutableLazyLoadConfiguration(..., base_path="base_path")  # Single Key
            MutableLazyLoadConfiguration(..., base_path="/base/path")  # JSON Pointer (strings only)
            MutableLazyLoadConfiguration(..., base_path=("base", "path"))  # List of keys

            # Use Environment Variable: "CONFIG_LOC"
            MutableLazyLoadConfiguration(..., env_location_var_name="CONFIG_LOC")

            # Use default Environment Variable: "G_CONFIG_LOCATION"
            MutableLazyLoadConfiguration(..., use_env_location=True)

    :note:
        - The Environment Variable is read as a comma-delimited list of configuration path that will be appended to ``load_order_location`` list.
        - Setting the Environment Variable is always optional.
        - Setting ``use_env_location`` to :py:data:`True` is only required if you don't change ``env_location_var_name`` from its default value of ``G_CONFIG_LOCATION``.
    """

    def __init__(
        self,
        *load_order_location: PathOrStr,
        base_path: str | tabc.Sequence[str] | None = None,
        use_env_location: bool = False,
        env_location_var_name: str = "G_CONFIG_LOCATION",
    ) -> None:
        super().__init__(
            *load_order_location,
            base_path=base_path,
            use_env_location=use_env_location,
            env_location_var_name=env_location_var_name,
            inject_before=None,
            inject_after=None,
            disable_caching=True,
            _mutable_configuration=True,
        )

    @property
    def config(self) -> MutableConfiguration:
        """Load and fetch the configuration. Configuration is cached for subsequent calls.

        :note: Loading the configuration is thread-safe and locks while the configuration is loaded to prevent
            duplicative processing and data

        """
        return typ.cast(MutableConfiguration, super().config)

    def __delitem__(self, key: typ.Any) -> None:
        del self.config[key]

    def __setitem__(self, key: typ.Any, value: typ.Any) -> None:
        self.config[key] = value

    def as_typed(self, typed_base: type[C]) -> typ.NoReturn:
        """
        :py:meth:`as_typed` is not supported for :py:class:`MutableLazyLoadConfiguration`.
        Use :py:class:`LazyLoadConfiguration`.
        """
        raise NotImplementedError(
            "`as_typed` is not supported for `MutableLazyLoadConfiguration`. Use `LazyLoadConfiguration`."
        )



from __future__ import annotations


class DoesNotExist(ValueError):
    pass


class EnvironmentVaribleNotFound(KeyError):
    pass


class ErrorWhileLoadingFileOccurred(ValueError):
    pass


class ErrorWhileLoadingConfig(Exception):
    pass


class ErrorWhileLoadingTags(Exception):
    pass


class GetConfigReadBeforeSetException(Exception):
    pass


class InterpolationWarning(Warning):
    pass


class InterpolationSyntaxError(Exception):
    pass


class IniKeyExistAsANonMapping(Exception):
    pass


class IniTryToReplaceExistingKey(Exception):
    pass


class IniUnsupportedError(Exception):
    pass


class InvalidBasePathException(KeyError):
    pass


class IsNotAClass(ValueError):
    pass


class IsNotCallable(ValueError):
    pass


class JSONPathOnlyWorksOnMappings(Exception):
    pass


class JSONPathQueryFailed(KeyError):
    pass


class JSONPointerQueryFailed(KeyError):
    pass


class ParseEnvParsingError(Exception):
    pass


class ParsingTriedToCreateALoop(Exception):
    pass


class PlaceholderConfigurationError(Exception):
    pass


class RefMustStartFromRoot(Exception):
    pass


class ReservedFileExtension(Exception):
    pass



from __future__ import annotations

import collections.abc as tabc

from granular_configuration_language.yaml.classes import RT, LazyEval, LazyRoot, Root, Tag


class LazyEvalBasic(LazyEval[RT]):
    def __init__(self, tag: Tag, value: tabc.Callable[[], RT]) -> None:
        super().__init__(tag)
        self.__value = value

    def _run(self) -> RT:
        return self.__value()


class LazyEvalWithRoot(LazyEval[RT]):
    def __init__(self, tag: Tag, root: LazyRoot, value: tabc.Callable[[Root], RT]) -> None:
        super().__init__(tag)
        self.__value = value
        self.__lazy_root = root

    def _run(self) -> RT:
        return self.__value(self.__lazy_root.root)



# isort:skip_file
if __name__ == "__main__":

    import os
    import sys

    from granular_configuration_language.yaml._tags import handlers

    print(
        handlers.pretty(
            as_json=len(sys.argv) < 2,
            width=os.get_terminal_size().columns,
        )
    )



# isort:skip_file
import granular_configuration_language.yaml.classes  # Import first
from granular_configuration_language.yaml.classes import LazyEval, LazyRoot, Masked, Placeholder
from granular_configuration_language.yaml.load import loads



from __future__ import annotations

import collections.abc as tabc
import typing as typ
from os import PathLike

from granular_configuration_language import Configuration, LazyLoadConfiguration, MutableLazyLoadConfiguration
from granular_configuration_language._build import _merge
from granular_configuration_language.proxy import SafeConfigurationProxy
from granular_configuration_language.yaml import LazyEval
from granular_configuration_language.yaml.load import obj_pairs_func


def merge(
    configs: tabc.Iterable[Configuration | LazyLoadConfiguration | LazyEval | PathLike | typ.Any],
    *,
    mutable: bool = False,
) -> Configuration:
    """Merges the provided configurations into a single configuration.

    - Filters out non-:py:class:`.Configuration` objects.
    - Extracts :py:class:`.Configuration` from :py:class:`.LazyEval` and :py:class:`.LazyLoadConfiguration`.
    - Any :py:class:`os.PathLike` objects are loaded via individual :py:class:`.LazyLoadConfiguration` instances.

    .. caution::
        Don't use :py:func:`.merge` as a replacement for :py:class:`.LazyLoadConfiguration`. It is less efficient and creates Load Boundaries on each configuration.

    :param ~collections.abc.Iterable[Configuration | LazyLoadConfiguration | LazyEval | ~os.PathLike | ~typing.Any] configs: Configurations
        to be merged
    :param bool, optional mutable: If :py:data:`True`, :py:class:`.MutableConfiguration` is used, else
        :py:class:`.Configuration` is used. Defaults to :py:data:`False`.

    :returns: Merged configuration. Empty if nothing was mergeable.
    :rtype: Configuration

    """

    def configuration_only(
        configs: tabc.Iterable[Configuration | LazyLoadConfiguration | LazyEval | PathLike | typ.Any],
    ) -> tabc.Iterator[Configuration]:
        for config in configs:
            if isinstance(config, LazyEval):
                config = config.result

            if isinstance(config, Configuration):
                yield config
            elif isinstance(config, SafeConfigurationProxy):
                yield config.copy()
            elif isinstance(config, LazyLoadConfiguration):
                yield config.config
            elif isinstance(config, PathLike):
                if mutable:
                    yield MutableLazyLoadConfiguration(config).config
                else:
                    yield LazyLoadConfiguration(config).config

    configuration_type = obj_pairs_func(mutable)
    base_config = configuration_type()
    return _merge(configuration_type, base_config, configuration_only(configs))



from __future__ import annotations

import collections.abc as tabc
import inspect
import json
import os
import typing as typ
from collections import OrderedDict
from importlib import import_module
from importlib.metadata import entry_points
from importlib.util import resolve_name
from itertools import chain, filterfalse
from operator import attrgetter, itemgetter
from pathlib import Path
from pprint import pformat

from granular_configuration_language.exceptions import ErrorWhileLoadingTags
from granular_configuration_language.yaml.decorators._base import Tag, TagConstructor
from granular_configuration_language.yaml.decorators._tag_tracker import is_not_lazy, is_with_ref, is_without_ref

ModuleName = typ.NewType("ModuleName", str)
PluginName = typ.NewType("PluginName", str)


def is_TagConstructor(obj: typ.Any) -> typ.TypeGuard[TagConstructor]:
    return isinstance(obj, TagConstructor)


def get_tags_in_module(module_name: ModuleName) -> tabc.Iterator[TagConstructor]:
    return map(itemgetter(1), inspect.getmembers(import_module(module_name), is_TagConstructor))


def get_internal_tag_plugins() -> tabc.Iterator[ModuleName]:
    import granular_configuration_language.yaml._tags as tags

    tags_package = tags.__package__
    tags_module_path = Path(tags.__file__).parent
    private_sub_module_pattern = r"_[a-zA-Z]*.py"

    def get_abs_name(module_name: str) -> ModuleName:
        return ModuleName(resolve_name("." + module_name, package=tags_package))

    return map(
        get_abs_name,
        filter(
            None,
            map(
                inspect.getmodulename,
                tags_module_path.glob(private_sub_module_pattern),
            ),
        ),
    )


def get_external_tag_plugins() -> tabc.Iterator[tuple[PluginName, ModuleName]]:
    return map(attrgetter("name", "module"), entry_points(group="granular_configuration_language_20_tag"))


def get_all_tag_plugins(*, disable_plugin: typ.AbstractSet[str]) -> tabc.Iterator[ModuleName]:
    for module in get_internal_tag_plugins():
        yield module

    for name, module in get_external_tag_plugins():
        if name not in disable_plugin:
            yield module


class TagSet(tabc.Iterable[TagConstructor], typ.Container[str]):
    def __init__(self, tags: tabc.Iterable[TagConstructor]) -> None:
        self.__state: OrderedDict[Tag, TagConstructor] = OrderedDict()

        for tc in tags:
            tag = tc.tag
            if tag in self.__state:
                raise ErrorWhileLoadingTags(
                    f"Tag is already defined. `{repr(tc)}` attempted to replace `{repr(self.__state[tag])}`"
                )
            else:
                self.__state[tag] = tc

    def __contains__(self, x: typ.Any) -> bool:
        return x in self.__state

    def __iter__(self) -> tabc.Iterator[TagConstructor]:
        return iter(self.__state.values())

    def has_tags(self, *tags: Tag | str) -> bool:
        return all(map(self.__contains__, tags))

    def does_not_have_tags(self, *tags: Tag | str) -> bool:
        return not any(map(self.__contains__, tags))

    def __repr__(self) -> str:
        return f"TagSet{{{','.join(sorted(self.__state.keys()))}}}"

    def pretty(self, as_json: bool = False, width: int = 120, indent: int = 2) -> str:
        if as_json:

            def get_type(tc: TagConstructor) -> typ.Any:
                attributes: list[str] = list()

                if is_with_ref(tc.constructor):
                    attributes.append("interpolates")
                elif is_without_ref(tc.constructor):
                    attributes.append("interpolates-reduced")

                if is_not_lazy(tc.constructor):
                    attributes.append("NOT-LAZY")

                return dict(
                    input=tc.friendly_type,
                    output=repr(inspect.signature(tc.constructor).return_annotation),
                    notes=attributes,
                )

        else:

            def get_type(tc: TagConstructor) -> typ.Any:
                attributes: set[str] = set()

                if is_with_ref(tc.constructor):
                    attributes.add("interpolates")
                elif is_without_ref(tc.constructor):
                    attributes.add("interpolates-reduced")

                if is_not_lazy(tc.constructor):
                    attributes.add("NOT-LAZY")

                return (
                    tc.friendly_type
                    + (f" [{', '.join(attributes)}]" if attributes else "")
                    + f" ({tc.constructor.__module__}.{tc.constructor.__name__})"
                    + f" `{inspect.signature(tc.constructor).return_annotation}`"
                )

        tags = {tag: get_type(tc) for tag, tc in self.__state.items()}

        if as_json:
            body = json.dumps(tags, indent=indent, sort_keys=True)[3:-2]
        else:
            body = pformat(tags, width=width, indent=indent, sort_dicts=True)[1:-1]

        return f"""TagSet{{\n {body}\n}}"""


def load_tags(
    *,
    disable_plugins: typ.AbstractSet[str] = frozenset(),
    disable_tags: typ.AbstractSet[Tag | str] = frozenset(),
) -> TagSet:
    disable_plugins |= frozenset(filter(None, map(str.strip, os.getenv("G_CONFIG_DISABLE_PLUGINS", "").split(","))))
    disable_tags |= frozenset(filter(None, map(str.strip, os.getenv("G_CONFIG_DISABLE_TAGS", "").split(","))))

    return TagSet(
        filterfalse(
            disable_tags.__contains__,
            chain.from_iterable(map(get_tags_in_module, get_all_tag_plugins(disable_plugin=disable_plugins))),
        )
    )



from __future__ import annotations

import collections.abc as tabc
import dataclasses
import operator as op
import typing as typ
from collections import deque
from functools import cached_property, reduce
from threading import Lock
from weakref import WeakValueDictionary

from granular_configuration_language._base_path import BasePath, read_base_path
from granular_configuration_language._build import build_configuration
from granular_configuration_language._configuration import Configuration
from granular_configuration_language._locations import Locations


@dataclasses.dataclass(frozen=False, eq=False, kw_only=True)
class SharedConfigurationReference:
    _locations: Locations
    _mutable_config: bool
    _inject_before: Configuration | None = None
    _inject_after: Configuration | None = None
    __lock: Lock | None = dataclasses.field(repr=False, compare=False, init=False, default_factory=Lock)
    __notes: deque[NoteOfIntentToRead] = dataclasses.field(repr=False, compare=False, init=False, default_factory=deque)

    def register(self, note: NoteOfIntentToRead) -> None:
        self.__notes.append(note)

    def __clear_notes(self, caller: NoteOfIntentToRead) -> None:
        while self.__notes:
            note = self.__notes.pop()
            if note is not caller:
                note._config

    def build(self, caller: NoteOfIntentToRead) -> Configuration:
        # Making cached_property thread-safe
        if self.__lock:
            with self.__lock:
                self.__config
                self.__lock = None
                self.__clear_notes(caller)

        return self.__config

    @cached_property
    def __config(self) -> Configuration:
        return build_configuration(
            self._locations, self._mutable_config, inject_after=self._inject_after, inject_before=self._inject_before
        )


@dataclasses.dataclass(frozen=False, eq=False, kw_only=True)
class NoteOfIntentToRead:
    _base_path: BasePath
    _config_ref: SharedConfigurationReference

    def __post_init__(self) -> None:
        self._config_ref.register(self)

    @property
    def config(self) -> Configuration:
        config = self._config
        if isinstance(config, Exception):
            raise config
        else:
            return config

    @cached_property
    def _config(self) -> Configuration | Exception:
        config = self._config_ref.build(self)
        try:
            return reduce(op.getitem, self._base_path, config)
        except Exception as e:
            return e
        finally:
            del self._config_ref


store: typ.Final[WeakValueDictionary[Locations, SharedConfigurationReference]] = WeakValueDictionary()


def prepare_to_load_configuration(
    *,
    locations: Locations,
    base_path: str | tabc.Sequence[str] | None,
    mutable_configuration: bool,
    disable_cache: bool,
    inject_before: Configuration | None,
    inject_after: Configuration | None,
) -> NoteOfIntentToRead:
    if disable_cache or mutable_configuration or inject_after or inject_before:
        shared_config_ref = SharedConfigurationReference(
            _locations=locations,
            _mutable_config=mutable_configuration,
            _inject_after=inject_after,
            _inject_before=inject_before,
        )
    elif locations not in store:
        shared_config_ref = SharedConfigurationReference(_locations=locations, _mutable_config=mutable_configuration)
        store[locations] = shared_config_ref
    else:
        shared_config_ref = store[locations]

    return NoteOfIntentToRead(_base_path=read_base_path(base_path), _config_ref=shared_config_ref)



from __future__ import annotations

from granular_configuration_language._configuration import Configuration
from granular_configuration_language.yaml.decorators import Tag, as_lazy, mapping_of_any_tag


@mapping_of_any_tag(Tag("!Dict"))
@as_lazy
def handler(value: Configuration) -> dict:
    return value.as_dict()



from __future__ import annotations

import re
import typing as typ

from granular_configuration_language._utils import get_environment_variable
from granular_configuration_language.yaml.decorators import Tag, as_lazy, string_tag

ENV_PATTERN: typ.Pattern[str] = re.compile(r"(\{\{\s*(?P<env_name>[A-Za-z0-9-_]+)\s*(?:\:(?P<default>.*?))?\}\})")


def load_env(env_name: str, default: typ.Optional[str] = None) -> str:
    return get_environment_variable(env_name, default)


@string_tag(Tag("!Env"))
@as_lazy
def handler(value: str) -> str:
    return ENV_PATTERN.sub(lambda x: load_env(**x.groupdict()), value)



from __future__ import annotations

from granular_configuration_language.yaml.classes import Placeholder
from granular_configuration_language.yaml.decorators import Tag, as_not_lazy, string_tag


@string_tag(Tag("!Placeholder"))
@as_not_lazy
def handler(value: str) -> Placeholder:
    return Placeholder(value)



from __future__ import annotations

import typing as typ
from copy import copy
from functools import partial

from ruamel.yaml import YAML, SafeConstructor
from ruamel.yaml.resolver import BaseResolver

from granular_configuration_language.yaml._tags import handlers
from granular_configuration_language.yaml.classes import StateHolder
from granular_configuration_language.yaml.load._constructors import construct_mapping, construct_sequence


def make_constructor_class(state: StateHolder) -> typ.Type[SafeConstructor]:
    class ExtendedSafeConstructor(SafeConstructor):
        yaml_constructors = copy(SafeConstructor.yaml_constructors)

    for handler in handlers:
        handler(ExtendedSafeConstructor, state)

    ExtendedSafeConstructor.add_constructor(
        BaseResolver.DEFAULT_MAPPING_TAG, partial(construct_mapping, state.options.obj_pairs_func)
    )
    ExtendedSafeConstructor.add_constructor(
        BaseResolver.DEFAULT_SEQUENCE_TAG, partial(construct_sequence, state.options.sequence_func)
    )

    return ExtendedSafeConstructor


def load_yaml_string(config_str: str, state: StateHolder) -> typ.Any:
    if config_str.startswith("%YAML"):
        yaml = YAML(typ="rt")
    else:
        yaml = YAML(typ="safe")

    yaml.Constructor = make_constructor_class(state)
    return yaml.load(config_str)



from __future__ import annotations

import typing as typ


class ParsedEnvironmentVariableSyntax(typ.NamedTuple):
    mode: str | None
    name: slice
    value: slice
    next: int


default_slices = ParsedEnvironmentVariableSyntax("", slice(None, None), slice(None, None), 0)


def _get_mode(value: str, index: int) -> str | None:
    try:
        return value[index]
    except IndexError:
        return None


def get_parser_from_index(value: str, start_index: int) -> ParsedEnvironmentVariableSyntax:
    colon_index = value.find(":", start_index)

    if colon_index < 0:
        return default_slices
    else:
        return ParsedEnvironmentVariableSyntax(
            _get_mode(value, colon_index + 1),
            slice(None, colon_index),
            slice(colon_index + 2, None),
            colon_index + 2,
        )


def parse_environment_variable_syntax(contents: str) -> ParsedEnvironmentVariableSyntax:
    parser = get_parser_from_index(contents, 0)

    while parser.mode == ":":
        parser = get_parser_from_index(contents, parser.next)

    return parser



from __future__ import annotations

import collections.abc as tabc
import typing as typ

from granular_configuration_language._configuration import Configuration
from granular_configuration_language.yaml.decorators._base import TagDecoratorBase


class string_tag(TagDecoratorBase[str]):
    """A decorator factory for Tags that take a YAML string as argument.

    :example:
        .. code-block:: python

            @string_tag(Tag("!Tag"))
            @as_lazy
            def tag(value: str) -> Any:
                ...

    """

    Type: typ.TypeAlias = str
    """TypeAlias for this Tag factory"""

    @property
    def user_friendly_type(self) -> str:
        """"""  # Make undocumented
        return "str"

    def scalar_node_type_check(self, value: str) -> typ.TypeGuard[Type]:
        """"""  # Make undocumented
        return True


class string_or_twople_tag(TagDecoratorBase[str | tuple[str, typ.Any]]):
    """A decorator factory for Tags that take a YAML string or tuple of a YAML strings and YAML object as argument.

    :example:
        .. code-block:: python

            @string_or_twople_tag(Tag("!Tag"))
            @as_lazy
            def tag(value: string_or_twople_tag.Type) -> Any:
                ...

    """

    Type: typ.TypeAlias = str | tuple[str, typ.Any]
    """TypeAlias for this Tag factory"""

    @property
    def user_friendly_type(self) -> str:
        """"""  # Make undocumented
        return "str | tuple[str, Any]"

    def scalar_node_type_check(self, value: str) -> typ.TypeGuard[str]:
        """"""  # Make undocumented
        return True

    def sequence_node_type_check(self, value: tabc.Sequence) -> typ.TypeGuard[tuple[str, typ.Any]]:
        """"""  # Make undocumented
        return (1 <= len(value) <= 2) and isinstance(value[0], str)

    def sequence_node_transformer(self, value: typ.Any) -> Type:
        """"""  # Make undocumented
        if len(value) == 2:
            return value
        else:
            return value[0]


class sequence_of_any_tag(TagDecoratorBase[tabc.Sequence[typ.Any]]):
    """A decorator factory for Tags that take a YAML sequence as argument.

    :example:
        .. code-block:: python

            @sequence_of_any_tag(Tag("!Tag"))
            @as_lazy
            def tag(value: Sequence[Any]) -> Any:
                ...

    """

    Type: typ.TypeAlias = tabc.Sequence[typ.Any]
    """TypeAlias for this Tag factory"""

    @property
    def user_friendly_type(self) -> str:
        """"""  # Make undocumented
        return "list[Any]"

    def sequence_node_type_check(self, value: tabc.Sequence) -> typ.TypeGuard[Type]:
        """"""  # Make undocumented
        return True


class mapping_of_any_tag(TagDecoratorBase[Configuration]):
    """A decorator factory for Tags that take a YAML mapping as argument.

    :example:
        .. code-block:: python

            @mapping_of_any_tag(Tag("!Tag"))
            @as_lazy
            def tag(value: Configuration) -> Any:
                ...

    """

    Type: typ.TypeAlias = Configuration
    """TypeAlias for this Tag factory"""

    @property
    def user_friendly_type(self) -> str:
        """"""  # Make undocumented
        return "dict[Any, Any]"

    def mapping_node_type_check(self, value: tabc.Mapping) -> typ.TypeGuard[Type]:
        """"""  # Make undocumented
        return True



from __future__ import annotations

import collections.abc as tabc
import typing as typ

from ruamel.yaml import MappingNode, SafeConstructor, SequenceNode

from granular_configuration_language.yaml.classes import LazyEval


def construct_mapping(cls: typ.Type, constructor: SafeConstructor, node: MappingNode) -> tabc.Mapping:
    node.value = [pair for pair in node.value if pair[0].tag != "!Del"]

    value: tabc.Mapping = cls(constructor.construct_mapping(node, deep=False))

    for key in value.keys():
        if isinstance(key, LazyEval):
            raise TypeError("Lazy Tags are not allowed as keys to mappings.")

    return cls(constructor.construct_mapping(node, deep=False))


def construct_sequence(cls: typ.Type, constructor: SafeConstructor, node: SequenceNode) -> tabc.Sequence:
    value = constructor.construct_sequence(node, deep=False)

    if isinstance(value, cls):
        return value
    else:
        return cls(value)



from __future__ import annotations

import collections.abc as tabc
import typing as typ
from functools import partial
from itertools import chain
from pathlib import Path

from granular_configuration_language import Configuration
from granular_configuration_language._load import load_file
from granular_configuration_language._s import setter_secret
from granular_configuration_language._utils import consume
from granular_configuration_language.yaml import LazyRoot
from granular_configuration_language.yaml.load import obj_pairs_func

_C = typ.TypeVar("_C", bound=Configuration)


def _merge_into_base(configuration_type: typ.Type[_C], base_dict: _C, from_dict: _C) -> None:
    for key, value in from_dict._raw_items():
        if isinstance(value, configuration_type) and (key in base_dict):
            if base_dict.exists(key):
                new_dict = base_dict[key]
            else:  # If Placeholder
                new_dict = configuration_type()

            if isinstance(new_dict, configuration_type):
                _merge_into_base(configuration_type, new_dict, value)
                value = new_dict

        base_dict._private_set(key, value, setter_secret)


def _merge(configuration_type: typ.Type[_C], base_config: _C, configs: tabc.Iterable[_C]) -> _C:
    consume(map(partial(_merge_into_base, configuration_type, base_config), configs))
    return base_config


def _load_configs_from_locations(
    configuration_type: typ.Type[_C], locations: tabc.Iterable[Path], lazy_root: LazyRoot, mutable: bool
) -> tabc.Iterator[_C]:
    def configuration_only(
        configs: tabc.Iterable[_C | typ.Any],
    ) -> tabc.Iterator[_C]:
        for config in configs:
            if isinstance(config, configuration_type):
                yield config

    _load_file = partial(load_file, lazy_root=lazy_root, mutable=mutable)
    return configuration_only(map(_load_file, locations))


def _inject(*new: Configuration, configs: tabc.Iterator[_C], in_front: bool) -> tabc.Iterator[_C]:
    cast = typ.cast(_C, new)
    if in_front:
        return chain(cast, configs)
    else:
        return chain(configs, cast)


def _inject_configs(
    configs: tabc.Iterator[_C],
    *,
    before: Configuration | None,
    after: Configuration | None,
) -> tabc.Iterator[_C]:
    if before and isinstance(before, Configuration):
        configs = _inject(before, configs=configs, in_front=True)

    if after and isinstance(after, Configuration):
        configs = _inject(after, configs=configs, in_front=False)

    return configs


def build_configuration(
    locations: tabc.Iterable[Path],
    mutable: bool,
    *,
    inject_before: Configuration | None,
    inject_after: Configuration | None,
) -> Configuration:
    configuration_type = obj_pairs_func(mutable)
    base_config = configuration_type()
    lazy_root = LazyRoot.with_root(base_config)

    valid_configs = _inject_configs(
        _load_configs_from_locations(configuration_type, locations, lazy_root, mutable),
        before=inject_before,
        after=inject_after,
    )
    merged_config = _merge(configuration_type, base_config, valid_configs)

    return merged_config



from __future__ import annotations

import collections.abc as tabc
import operator as op
import re
import typing as typ

import jsonpath

from granular_configuration_language.exceptions import (
    JSONPathOnlyWorksOnMappings,
    JSONPathQueryFailed,
    JSONPointerQueryFailed,
    RefMustStartFromRoot,
)
from granular_configuration_language.yaml.classes import LazyEval
from granular_configuration_language.yaml.decorators import Root

SUB_PATTERN: typ.Pattern[str] = re.compile(r"(\$\{(?P<contents>.*?)\})")


def _resolve_pointer(query: str, root: tabc.Mapping) -> typ.Any:
    try:
        not_found = object()

        result = jsonpath.JSONPointer(query).resolve(root, default=not_found)

        if result is not_found:
            raise JSONPointerQueryFailed(f"JSON Pointer `{query}` did not find a match.")
        else:
            return result

    except RecursionError:
        raise RecursionError(
            (
                f"JSON Pointer `{query}` caused a recursion error. Please check your configuration for a self-referencing loop."
            )
        ) from None


def _resolve_path(query: str, root: tabc.Mapping) -> typ.Any:
    try:
        result = list(map(op.attrgetter("value"), jsonpath.finditer(query, root)))

        if len(result) == 1:
            return result[0]
        elif len(result) == 0:
            raise JSONPathQueryFailed(f"JSON Path `{query}` did not find a match.")
        else:
            return result

    except RecursionError:
        raise RecursionError(
            (
                f"JSON Path `{query}` caused a recursion error. Please check your configuration for a self-referencing loop."
            )
        ) from None


def resolve_json_ref(query: str, root: Root) -> typ.Any:
    if isinstance(root, LazyEval) and root.tag == "!Merge":
        raise RecursionError(
            f"JSON Query `{query}` attempted recursion. Please check your configuration for a self-referencing loop."
        )
    elif not isinstance(root, tabc.Mapping):
        raise JSONPathOnlyWorksOnMappings(f"JSONPath `{query}` was tried on `{repr(root)}`")
    elif query.startswith("$"):
        return _resolve_path(query, root)
    elif query.startswith("/"):
        return _resolve_pointer(query, root)
    else:
        raise RefMustStartFromRoot(f"JSON query `{query}` must start with '$' for JSON Path or '/' for JSON Pointer")



from __future__ import annotations

from uuid import UUID

from granular_configuration_language.yaml.decorators import Tag, as_lazy, interpolate_value_without_ref, string_tag


@string_tag(Tag("!UUID"))
@as_lazy
@interpolate_value_without_ref
def handler(value: str) -> UUID:
    return UUID(hex=value)



from __future__ import annotations

import collections.abc as tabc
import operator as op
import os
import typing as typ
from functools import cached_property, lru_cache
from itertools import chain, islice
from pathlib import Path

from granular_configuration_language._utils import OrderedSet

PathOrStr = Path | str | os.PathLike


@lru_cache()
def _resolve_path(path: Path) -> Path:
    return path.expanduser().resolve()


def _convert_to_path(path: PathOrStr) -> Path:
    if isinstance(path, Path):
        return path
    else:
        return Path(path)


def path_repr(path: Path) -> str:  # pragma: no cover
    return str(path.relative_to(Path.cwd()))


class BaseLocation(tabc.Iterable[Path], typ.Hashable):
    pass


class PrioritizedLocations(BaseLocation):
    __slots__ = ("paths",)

    def __init__(self, paths: tuple[Path, ...]) -> None:
        self.paths: typ.Final = paths

    def __iter__(self) -> tabc.Iterator[Path]:
        return islice(filter(op.methodcaller("is_file"), self.paths), 1)

    def __eq__(self, value: object) -> bool:
        return isinstance(value, PrioritizedLocations) and self.paths == value.paths

    def __hash__(self) -> int:
        return self.__hash

    @cached_property
    def __hash(self) -> int:
        return hash(self.paths)

    def __repr__(self) -> str:
        return f"<PrioritizedLocations=[{','.join(map(path_repr, self.paths))}]>"


class Location(BaseLocation):
    __slots__ = "path"

    def __init__(self, path: Path) -> None:
        self.path: typ.Final = path

    def __iter__(self) -> tabc.Iterator[Path]:
        if self.path.is_file():
            yield self.path

    def __eq__(self, value: object) -> bool:
        return isinstance(value, Location) and self.path == value.path

    def __hash__(self) -> int:
        return self.__hash

    @cached_property
    def __hash(self) -> int:
        return hash(self.path)

    def __repr__(self) -> str:
        return f"<Location={path_repr(self.path)}>"


SUFFIX_CONFIG: typ.Final[dict[str, tabc.Sequence[str]]] = {
    ".*": (".yaml", ".yml"),
    ".y*": (".yaml", ".yml"),
    ".yml": (".yaml", ".yml"),
}


@lru_cache()
def _convert_to_location(path: Path) -> BaseLocation:
    if path.suffix in SUFFIX_CONFIG:
        return PrioritizedLocations(tuple(map(path.with_suffix, SUFFIX_CONFIG[path.suffix])))
    else:
        return Location(path)


class Locations(BaseLocation):
    def __init__(self, locations: tabc.Iterable[PathOrStr]) -> None:
        self.locations: typ.Final = tuple(
            map(_convert_to_location, map(_resolve_path, map(_convert_to_path, locations)))
        )

    def __iter__(self) -> tabc.Iterator[Path]:
        return iter(OrderedSet(chain.from_iterable(self.locations)))

    def __bool__(self) -> bool:
        return bool(self.locations)

    def __eq__(self, value: object) -> bool:
        return isinstance(value, Locations) and self.locations == value.locations

    def __hash__(self) -> int:
        return self.__hash

    @cached_property
    def __hash(self) -> int:
        return sum(map(hash, self.locations))

    def __repr__(self) -> str:
        return f"<Locations=[{','.join(map(repr, self.locations))}]>"



from __future__ import annotations

import collections.abc as tabc
import os
import typing as typ
from functools import partial

from granular_configuration_language.exceptions import (
    EnvironmentVaribleNotFound,
    ParseEnvParsingError,
    ParsingTriedToCreateALoop,
)
from granular_configuration_language.yaml._parsing import (
    create_environment_variable_path,
    is_in_chain,
    make_chain_message,
)
from granular_configuration_language.yaml.classes import LazyRoot
from granular_configuration_language.yaml.decorators import (
    LoadOptions,
    Root,
    Tag,
    as_lazy,
    as_lazy_with_root_and_load_options,
    string_or_twople_tag,
)


def parse_env(load: tabc.Callable[[str, str], typ.Any], env_var: str, *default: typ.Any) -> typ.Any:
    env_missing = env_var not in os.environ

    if env_missing and (len(default) > 0):
        return default[0]
    elif env_missing:
        raise EnvironmentVaribleNotFound(env_var)
    else:
        try:
            return load(env_var, os.environ[env_var])
        except ParsingTriedToCreateALoop:
            raise
        except Exception as e:
            raise ParseEnvParsingError(
                f"Error while parsing Environment Variable ({env_var}): ({e.__class__.__name__}) {e}"
            ) from None


def load_advance(options: LoadOptions, root: Root, env_var: str, value: str) -> typ.Any:
    from granular_configuration_language.yaml import loads

    file_path = create_environment_variable_path(env_var)

    if is_in_chain(file_path, options):
        raise make_chain_message("!ParseEnv", env_var, options)

    lazy_root = LazyRoot.with_root(root)
    return loads(
        value,
        lazy_root=lazy_root,
        mutable=options.mutable,
        previous_options=options,
        file_path=file_path,
    )


def load_safe(env_var: str, value: str) -> typ.Any:
    from ruamel.yaml import YAML

    return YAML(typ="safe").load(value)


def parse_input(load: tabc.Callable[[str, str], typ.Any], value: string_or_twople_tag.Type) -> typ.Any:
    if isinstance(value, str):
        return parse_env(load, value)
    else:
        return parse_env(load, *value)


@string_or_twople_tag(Tag("!ParseEnv"))
@as_lazy_with_root_and_load_options
def handler(value: string_or_twople_tag.Type, root: Root, options: LoadOptions) -> typ.Any:
    return parse_input(partial(load_advance, options, root), value)


@string_or_twople_tag(Tag("!ParseEnvSafe"))
@as_lazy
def handler_safe(value: string_or_twople_tag.Type) -> typ.Any:
    return parse_input(load_safe, value)



from __future__ import annotations

from granular_configuration_language.yaml.classes import Masked
from granular_configuration_language.yaml.decorators import Tag, as_not_lazy, interpolate_value_without_ref, string_tag


@string_tag(Tag("!Mask"))
@as_not_lazy
@interpolate_value_without_ref
def handler(value: str) -> Masked:
    return Masked(value)



from __future__ import annotations

import collections.abc as tabc


class BasePathPart(str):
    pass


class BasePath(tuple[BasePathPart]):
    pass


def read_base_path(base_path: str | tabc.Sequence[str] | None) -> BasePath:
    if isinstance(base_path, str):
        if base_path.startswith("/"):
            return BasePath(map(BasePathPart, filter(None, base_path.split("/"))))
        else:
            return BasePath((BasePathPart(base_path),))
    elif base_path:
        return BasePath(map(BasePathPart, base_path))
    else:
        return BasePath()



# isort:skip_file
import granular_configuration_language.yaml.classes  # Needs to import before _configuration to prevent circular import
from granular_configuration_language._configuration import Configuration, MutableConfiguration
from granular_configuration_language._lazy_load_configuration import (
    LazyLoadConfiguration,
    LazyLoadConfiguration as LLC,
    MutableLazyLoadConfiguration,
)
from granular_configuration_language._merge import merge  # depends on LazyLoadConfiguration

from granular_configuration_language._json import json_default
from granular_configuration_language.yaml import Masked, Placeholder



from __future__ import annotations

import collections.abc as tabc
import os
import typing as typ
from collections import OrderedDict, deque
from functools import partial

from granular_configuration_language.exceptions import EnvironmentVaribleNotFound

consume = typ.cast(tabc.Callable[[tabc.Iterable], None], partial(deque, maxlen=0))

_T = typ.TypeVar("_T")


class OrderedSet(tabc.MutableSet[_T], typ.Reversible[_T], typ.Generic[_T]):
    def __init__(self, iterable: typ.Optional[tabc.Iterable[_T]] = None) -> None:
        self.__backend: OrderedDict[_T, None] = OrderedDict()
        if iterable is not None:
            consume(map(self.add, iterable))

    def __contains__(self, item: typ.Any) -> bool:
        return item in self.__backend

    def __iter__(self) -> tabc.Iterator[_T]:
        return iter(self.__backend)

    def __len__(self) -> int:
        return len(self.__backend)

    def add(self, value: _T) -> None:
        self.__backend[value] = None

    def discard(self, value: _T) -> None:
        del self.__backend[value]

    def __reversed__(self) -> tabc.Iterator[_T]:
        return reversed(self.__backend)


def get_environment_variable(name: str, default: str | tabc.Callable[[], str] | None = None) -> str:
    name = name.replace("::", ":")
    if name in os.environ:
        return os.environ[name]
    elif default is None:
        raise EnvironmentVaribleNotFound(name)
    elif callable(default):
        return default()
    else:
        return default



from __future__ import annotations

import collections.abc as tabc
import re
import typing as typ
import warnings
from functools import wraps
from html import unescape

from granular_configuration_language._utils import get_environment_variable
from granular_configuration_language.exceptions import InterpolationSyntaxError, InterpolationWarning
from granular_configuration_language.yaml.classes import RT, P, Root
from granular_configuration_language.yaml.decorators._tag_tracker import track_as_with_ref, track_as_without_ref
from granular_configuration_language.yaml.decorators.interpolate._env_var_parser import (
    parse_environment_variable_syntax,
)
from granular_configuration_language.yaml.decorators.ref import resolve_json_ref


def _get_ref_string(root: Root, contents: str) -> str:
    value = resolve_json_ref(contents, root)
    if isinstance(value, str):
        return value
    elif isinstance(value, (tabc.Mapping, tabc.Sequence)):
        return repr(value)
    else:
        return str(value)


def _get_env_var_string(root: Root, contents: str) -> str:
    parser = parse_environment_variable_syntax(contents)
    match parser.mode:
        case "":
            return get_environment_variable(contents[parser.name])
        case "-":
            return get_environment_variable(contents[parser.name], contents[parser.value])
        case "+":
            return get_environment_variable(
                contents[parser.name], lambda: curly_sub(root, contents=contents[parser.value])
            )
        case _:
            raise InterpolationSyntaxError(
                f'":{parser.mode}" is not a supported environment variable interpolation mode.'
            )


def curly_sub(root: Root, *, contents: str) -> str:
    if contents == "":
        raise InterpolationSyntaxError(
            'Empty expression ("${}" or "${...:+}") is not a supported environment variable interpolation syntax.'
        )
    elif contents == "$":
        return "$"
    elif root and contents.startswith("$") or contents.startswith("/"):
        return _get_ref_string(root, contents)
    elif contents.startswith("&") and contents.endswith(";"):
        return unescape(contents)
    else:
        return _get_env_var_string(root, contents)


def round_sub(root: Root, *, contents: str) -> str:
    warnings.warn("`!Sub $()` is reserved", InterpolationWarning)
    return "$(" + contents + ")"


SUB_PATTERNS: typ.Final[tabc.Sequence[tuple[tabc.Callable, typ.Pattern[str]]]] = (
    (round_sub, re.compile(r"(\$\((?P<contents>.*?)\))")),
    (curly_sub, re.compile(r"(\$\{(?P<contents>.*?)\})")),
)


def interpolate(value: str, root: Root) -> str:
    for sub, pat in SUB_PATTERNS:
        value = pat.sub(lambda x: sub(root, **x.groupdict()), value)
    return value


# Trying to explain with variable names
DOLLAR_BUT_NOT_END = r"\$(?!\})"
SLASH = r"/"
DOLLAR_OR_SLASH = f"(?:{DOLLAR_BUT_NOT_END}|{SLASH})"
NESTING = r".+?\:\+"
NESTING_DOLLAR_OR_SLASH = NESTING + DOLLAR_OR_SLASH
STARTS_WITH_OR_NESTING_DOLLAR_OR_SLASH = f"(?:{DOLLAR_OR_SLASH}|{NESTING_DOLLAR_OR_SLASH})"
DOLLAR_BRACKET = r"\$\{"
WHOLE_THING = DOLLAR_BRACKET + STARTS_WITH_OR_NESTING_DOLLAR_OR_SLASH
# WHOLE_THING = r"\$\{(?:(?:\$(?!\})|/)|.+?\:\+(?:\$(?!\})|/))"

DOES_REF_PATTERN = re.compile(WHOLE_THING)

del (
    DOLLAR_BUT_NOT_END,
    SLASH,
    DOLLAR_OR_SLASH,
    NESTING,
    NESTING_DOLLAR_OR_SLASH,
    STARTS_WITH_OR_NESTING_DOLLAR_OR_SLASH,
    DOLLAR_BRACKET,
    WHOLE_THING,
)  # Clean up all those temporary variables


def interpolation_needs_ref_condition(value: str) -> bool:
    return bool(DOES_REF_PATTERN.search(value))


def interpolate_value_with_ref(
    func: tabc.Callable[typ.Concatenate[str, Root, P], RT],
) -> tabc.Callable[typ.Concatenate[str, Root, P], RT]:
    """Replaces the YAML string value with the interpolated value before calling the tag function

    "with_ref" does full interpolation, supporting references (e.g. ``${$.value}`` and ``${/value}``).

    :param ~collections.abc.Callable[~typing.Concatenate[str, ~granular_configuration_language.yaml.classes.Root, P], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[~typing.Concatenate[str, ~granular_configuration_language.yaml.classes.Root, P], RT]

    :note: - First positional argument must a :py:class:`str`.
        - Second positional must be :py:type:`~granular_configuration_language.yaml.classes.Root` type, even if you do not use it.
    :example:
        .. code-block:: python

            @string_tag(Tag("!Tag"))
            @as_lazy_with_root
            @interpolate_value_with_ref
            def tag(value: str, root: Root) -> Any:
                ...

            @string_tag(Tag("!Tag"))
            @as_lazy_with_root_and_load_options
            @interpolate_value_with_ref
            def tag_with_options(value: str, root: Root, options: LoadOptions) -> Any:
                ...
    """

    @wraps(func)
    def lazy_wrapper(value: str, root: Root, /, *args: P.args, **kwargs: P.kwargs) -> RT:
        return func(interpolate(value, root), root, *args, **kwargs)

    track_as_with_ref(func)
    return lazy_wrapper


def interpolate_value_without_ref(
    func: tabc.Callable[typ.Concatenate[str, P], RT],
) -> tabc.Callable[typ.Concatenate[str, P], RT]:
    """Replaces the YAML string value with the interpolated value before calling the tag function

    "without_ref" does a limited interpolation that does not support references (e.g. ``${$.value}`` and ``${/value}``)

    :param ~collections.abc.Callable[~typing.Concatenate[str, P], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[~typing.Concatenate[str, P], RT]

    :note: - First positional argument must a :py:class:`str`.
    :example:
        .. code-block:: python

            @string_tag(Tag("!Tag"))
            @as_lazy
            @interpolate_value_with_ref
            def tag(value: str) -> Any:
                ...


            @string_tag(Tag("!Tag"))
            @as_lazy_with_load_options
            @interpolate_value_with_ref
            def tag_with_options(value: str, options: LoadOptions) -> Any:
                ...

    """

    @wraps(func)
    def lazy_wrapper(value: str, /, *args: P.args, **kwargs: P.kwargs) -> RT:
        return func(interpolate(value, None), *args, **kwargs)

    track_as_without_ref(func)
    return lazy_wrapper



from __future__ import annotations

import collections.abc as tabc
import typing as typ
from functools import wraps

from granular_configuration_language.yaml.classes import RT, LazyEval, LoadOptions, Root, StateHolder, T, Tag
from granular_configuration_language.yaml.decorators._lazy_eval import LazyEvalBasic, LazyEvalWithRoot
from granular_configuration_language.yaml.decorators._tag_tracker import track_as_not_lazy


def as_lazy(func: tabc.Callable[[T], RT]) -> tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]:
    """Wraps the "Tag" function in a :py:class:`~.LazyEval`, so that the function being wrapped is run just-in-time.

    :param ~collections.abc.Callable[[T], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[[Tag, T, StateHolder], LazyEval[RT]]

    :note: - First Positional Argument: YAML value
    :example:
        .. code-block:: python

            @string_tag(Tag("!Tag"))
            @as_lazy
            def tag(value: str) -> Any:
                ...
    """

    @wraps(func)
    def lazy_wrapper(tag: Tag, value: T, state: StateHolder) -> LazyEvalBasic[RT]:
        return LazyEvalBasic(tag, lambda: func(value))

    return lazy_wrapper


def as_lazy_with_load_options(
    func: tabc.Callable[[T, LoadOptions], RT],
) -> tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]:
    """Wraps the "Tag" function in a :py:class:`~.LazyEval`, so that the function being wrapped is run just-in-time.

    :param ~collections.abc.Callable[[T, LoadOptions], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[[Tag, T, StateHolder], LazyEval[RT]]

    :note: - First Positional Argument: YAML value
        - Second Positional Argument: A :py:class:`.LoadOptions` instance
    :example:
        .. code-block:: python

            @string_tag(Tag("!Tag"))
            @as_lazy_with_load_options
            def tag(value: str, options: LoadOptions) -> Any:
                ...
    """

    @wraps(func)
    def lazy_wrapper(tag: Tag, value: T, state: StateHolder) -> LazyEvalBasic[RT]:
        options = state.options
        return LazyEvalBasic(tag, lambda: func(value, options))

    return lazy_wrapper


@typ.overload
def as_lazy_with_root(func: tabc.Callable[[T, Root], RT]) -> tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]: ...


@typ.overload
def as_lazy_with_root(
    *, needs_root_condition: tabc.Callable[[T], bool]
) -> tabc.Callable[[tabc.Callable[[T, Root], RT]], tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]]: ...


def as_lazy_with_root(
    func: tabc.Callable[[T, Root], RT] | None = None, *, needs_root_condition: tabc.Callable[[T], bool] | None = None
) -> (
    tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]
    | tabc.Callable[[tabc.Callable[[T, Root], RT]], tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]]
):
    r"""Wraps the "Tag" function in a :py:class:`~.LazyEval`, so that the function being wrapped is run just-in-time.

    Note:
        autodoc isn't exposing the :py:func:`typing.overload`. See the example for a clearer type signatures

    :param ~collections.abc.Callable[[T, Root], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[[Tag, T, StateHolder], LazyEval[RT]]

    :example:
        .. code-block:: python

            # Typical usage
            @string_tag(Tag("!Tag"))
            @as_lazy_with_root
            def tag(value: str, root: Root) -> Any:
                ...

            # Using `needs_root_condition`
            @string_tag(Tag("!Tag"))
            @as_lazy_with_root(needs_root_condition=interpolation_needs_ref_condition)
            @interpolate_value_with_ref
            def tag(value: str, root: Root) -> Any:
                ...
    """

    def decorator_generator(func: tabc.Callable[[T, Root], RT]) -> tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]:

        @wraps(func)
        def lazy_wrapper(tag: Tag, value: T, state: StateHolder) -> LazyEval[RT]:

            if (needs_root_condition is None) or needs_root_condition(value):
                return LazyEvalWithRoot(tag, state.lazy_root_obj, lambda root: func(value, root))
            else:
                return LazyEvalBasic(tag, lambda: func(value, None))

        return lazy_wrapper

    if func is None:
        return decorator_generator
    else:
        return decorator_generator(func)


def as_lazy_with_root_and_load_options(
    func: tabc.Callable[[T, Root, LoadOptions], RT],
) -> tabc.Callable[[Tag, T, StateHolder], LazyEval[RT]]:
    """Wraps the "Tag" function in a :py:class:`~.LazyEval`, so that the function being wrapped is run just-in-time.

    :param ~collections.abc.Callable[[T, ~granular_configuration_language.yaml.classes.Root, LoadOptions], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[[Tag, T, StateHolder], LazyEval[RT]]

    :note: - First Positional Argument: YAML value
        - Second Positional Argument: Configuration root, as type :py:type:`~granular_configuration_language.yaml.classes.Root`
        - Third Positional Argument: A :py:class:`.LoadOptions` instance
    :example:
        .. code-block:: python

            @string_tag(Tag("!Tag"))
            @as_lazy_with_root_and_load_options
            def tag(value: str, root: Root, options: LoadOptions) -> Any:
                ...
    """

    @wraps(func)
    def lazy_wrapper(tag: Tag, value: T, state: StateHolder) -> LazyEvalWithRoot[RT]:
        options = state.options
        return LazyEvalWithRoot(tag, state.lazy_root_obj, lambda root: func(value, root, options))

    return lazy_wrapper


def as_not_lazy(func: tabc.Callable[[T], RT]) -> tabc.Callable[[Tag, T, StateHolder], RT]:
    """Wraps the "Tag" function, but does not make it lazy. The function being wrapped is run at load time.

    :param ~collections.abc.Callable[[T], RT] func: Function to be wrapped

    :returns: Wrapped Function
    :rtype: ~collections.abc.Callable[[Tag, T, StateHolder], LazyEval[RT]]

    :note: - First Positional Argument: YAML value
    :example:
        .. code:: python

            @string_tag(Tag("!Tag"))
            @as_not_lazy
            def tag(value: str) -> Any:
                ...
    """

    @wraps(func)
    def lazy_wrapper(tag: Tag, value: T, state: StateHolder) -> RT:
        return func(value)

    track_as_not_lazy(func)
    return lazy_wrapper



from __future__ import annotations

import typing as typ
from pathlib import Path

from granular_configuration_language.yaml._parsing import is_in_chain, make_chain_message
from granular_configuration_language.yaml.classes import LazyRoot
from granular_configuration_language.yaml.decorators import (
    LoadOptions,
    Root,
    Tag,
    as_lazy_with_root_and_load_options,
    interpolate_value_with_ref,
    string_tag,
)


def _as_file_path(tag: str, value: str, options: LoadOptions) -> Path:
    result = options.relative_to_directory / value

    if is_in_chain(result, options):
        raise make_chain_message(tag, value, options)

    return result


def _load(file: Path, options: LoadOptions, root: Root) -> typ.Any:
    from granular_configuration_language._load import load_file

    lazy_root = LazyRoot.with_root(root)
    output = load_file(file, lazy_root=lazy_root, mutable=options.mutable, previous_options=options)
    return output


@string_tag(Tag("!ParseFile"))
@as_lazy_with_root_and_load_options
@interpolate_value_with_ref
def handler(value: str, root: Root, options: LoadOptions) -> typ.Any:
    file = _as_file_path("!ParseFile", value, options)

    return _load(file, options, root)


@string_tag(Tag("!OptionalParseFile"))
@as_lazy_with_root_and_load_options
@interpolate_value_with_ref
def handler_optional(value: str, root: Root, options: LoadOptions) -> typ.Any:
    file = _as_file_path("!OptionalParseFile", value, options)

    if file.exists():
        return _load(file, options, root)
    else:
        return options.obj_pairs_func()



from __future__ import annotations

import typing as typ
from pathlib import Path

from granular_configuration_language.exceptions import (
    ErrorWhileLoadingFileOccurred,
    IniUnsupportedError,
    ParsingTriedToCreateALoop,
    ReservedFileExtension,
)
from granular_configuration_language.yaml import LazyRoot, loads
from granular_configuration_language.yaml._parsing import FILE_EXTENSION
from granular_configuration_language.yaml.classes import LoadOptions


def load_file(
    filename: Path,
    *,
    mutable: bool,
    lazy_root: typ.Optional[LazyRoot] = None,
    previous_options: LoadOptions | None = None,
) -> typ.Any:
    try:
        if filename.suffix == ".ini":
            raise IniUnsupportedError("INI support has been removed")
        elif filename.suffix == FILE_EXTENSION:
            raise ReservedFileExtension(f"`{FILE_EXTENSION}` is a reserved internal file extension")
        else:
            loader = loads

        return loader(
            filename.read_text(),
            lazy_root=lazy_root,
            file_path=filename,
            mutable=mutable,
            previous_options=previous_options,
        )
    except (IniUnsupportedError, ParsingTriedToCreateALoop, ReservedFileExtension):
        raise
    except FileNotFoundError as e:
        raise FileNotFoundError(e) from None
    except Exception as e:
        raise ErrorWhileLoadingFileOccurred(f'Problem in file "{filename}": ({e.__class__.__name__}) {e}')
